import enum
import functools
from datetime import datetime
from datetime import timedelta
from typing import Callable
from typing import Iterable
from typing import List
from typing import Type
from typing import Union

import requests


class Impact(enum.IntFlag):
    HOLIDAY = enum.auto()
    LOW = enum.auto()
    MEDIUM = enum.auto()
    HIGH = enum.auto()


class Currency(enum.IntFlag):
    USD = enum.auto()
    EUR = enum.auto()
    JPY = enum.auto()
    GBP = enum.auto()
    CAD = enum.auto()
    AUD = enum.auto()
    CHF = enum.auto()
    CNY = enum.auto()
    NZD = enum.auto()
    SEK = enum.auto()
    BRL = enum.auto()
    KRW = enum.auto()
    HKD = enum.auto()
    SGD = enum.auto()
    MXN = enum.auto()
    ZAR = enum.auto()
    INR = enum.auto()


@functools.lru_cache
def _get_calendar_events(datetime_from: datetime,
                         datetime_to: datetime,
                         importance: int,
                         currencies: int,
                         ) -> List[dict]:
    url = "https://www.mql5.com/en/economic-calendar/content"
    headers = {"x-requested-with": "XMLHttpRequest"}
    time_format = "%Y-%m-%dT%H:%M:%S"
    data = {
        "date_mode" : 0,
        "from"      : datetime_from.strftime(time_format),
        "to"        : datetime_to.strftime(time_format),
        "importance": importance,
        "currencies": currencies,
    }
    events = requests.post(url=url, headers=headers, data=data).json()
    return events


def _round_time_to_mins(time: datetime, round_minutes: int):
    secs = round_minutes * 60.0
    res = round(time.timestamp() / secs, 0) * secs
    new_time = datetime.fromtimestamp(res)
    return new_time


def _split_pairs(p: Iterable[str]):
    for s in p:
        yield from (s,) if len(s) < 6 else (s[:3], s[3:6])


@functools.lru_cache
def _make_flag(enum_cls: Union[Type[Impact], Type[Currency]],
               flags: Union[Iterable[str], int, str] = None
               ) -> int:
    if isinstance(flags, int):
        return int(flags)
    if flags is None:
        flags = enum_cls.__members__.keys()
    elif isinstance(flags, str):
        flags = flags.replace(',', ' ').split()
    if enum_cls is Currency:
        flags = _split_pairs(flags)
    flags = [enum_cls.__members__.get(f, 0) for f in set(map(str.upper, flags))]
    flag = functools.reduce(lambda x, y: x | y, flags)
    return int(flag)


def calendar_events(time_to: Union[datetime, timedelta] = None,
                    *,
                    time_from: Union[datetime, timedelta] = None,
                    importance: Union[Iterable[str], str, int] = None,
                    currencies: Union[Iterable[str], str, int] = None,
                    function: Callable = None,
                    round_minutes: int = 60,
                    cache_clear: bool = False,
                    **kwargs,
                    ) -> List[dict]:
    if cache_clear:
        _get_calendar_events.cache_clear()
    now = datetime.now()
    if time_to is None and time_from is None:
        time_to = now + timedelta(weeks=1)
    time_from, time_to = time_from or now, time_to or now
    _f = lambda x: now + x if isinstance(x, timedelta) else x
    time_from, time_to = _f(time_from), _f(time_to)
    if time_from > time_to:
        time_from, time_to = time_to, time_from
    time_from = _round_time_to_mins(time_from, round_minutes)
    time_to = _round_time_to_mins(time_to, round_minutes)
    _f = lambda x: tuple(x) if isinstance(x, Iterable) and not isinstance(x, str) else x
    importance, currencies = _f(importance), _f(currencies)
    i_flag, c_flag = _make_flag(Impact, importance), _make_flag(Currency, currencies)
    events = _get_calendar_events(datetime_from=time_from, datetime_to=time_to, importance=i_flag, currencies=c_flag)
    if function:
        events = list(filter(function, events))
    return events




from datetime import datetime

from .types import *


def as_dict_all(data: Any):
    """Convert all nested data returns to native python (pickleable) data structures. Example: List[OrderSendResult]
    -> List[dict]

    :param data: Any API returned result from the MetaTrader5 API
    :return:
    """
    if hasattr(data, '_asdict'):  # this is about twice as fast as try, except!
        return as_dict_all(data._asdict())
    T = type(data)
    if T is tuple or T is list:
        return T(as_dict_all(i) for i in data)
    if T is dict:
        return {k: as_dict_all(v) for k, v in data.items()}
    return data
    # try:
    #     return as_dict_all(data._asdict())
    # except AttributeError:
    #     T = type(data)
    #     if T is tuple or T is list:
    #         return T(as_dict_all(i) for i in data)
    #     if T is dict:
    #         return {k: as_dict_all(v) for k, v in data.items()}
    #     return data


def any_symbol(symbol):
    """Pass any symbol object with a name property or string.

    :param symbol: Any symbol.
    :return: Symbol as string.
    """
    try:
        return symbol.name
    except AttributeError:
        return symbol


def reduce_args(kwargs: dict) -> dict:
    return {k: v for k, v in kwargs.items() if v is not None and k != 'kwargs'}


def reduce_args_by_keys(d: dict, keys: Iterable) -> dict:
    return {k: v for k, v in d.items() if k in keys and v is not None}


def reduce_combine(d1: dict, d2: dict):
    d1 = reduce_args(d1)
    for k, v in d2.items():
        if v is not None:
            d1[k] = v
    return d1


def args_to_str(args: tuple, kwargs: dict):
    ar = ', '.join(map(str, args))
    kw = ', '.join(f"{k}={v}" for k, v in kwargs.items())
    return ar + (', ' if ar and kw else '') + kw


def is_rates_array(array):
    try:
        rate = array[0]
        return type(rate) is tuple and len(rate) == 8
    except:
        return False


def get_ticket_type_stuff(func, *, symbol, group, ticket, function):
    d = locals().copy()
    kw = reduce_args_by_keys(d, ['symbol', 'group', 'ticket'])
    items = func(**kw)
    # if magic:
    #     items = filter(lambda p: p.magic == magic, items)
    if function:
        items = filter(function, items)
    return tuple(items) if items is not None else tuple()


def get_history_type_stuff(func, args):
    args = reduce_args(args)
    function = args.pop('function', None)
    datetime_from = args.get('datetime_from', None)
    datetime_to = args.get('datetime_to', None)
    if not args:
        datetime_from = datetime(2000, 1, 1)
        datetime_to = datetime.now()
    if datetime_from is not None and datetime_to is not None:
        deals = func(datetime_from, datetime_to, **args)
    else:
        deals = func(**args)
    if function:
        deals = tuple(filter(function, deals))
    return deals if deals is not None else tuple()


def do_trade_action(func, args):
    cleaned = reduce_args(args)
    request = cleaned.pop('request', {})
    symbol = cleaned.pop('symbol', None) or request.pop('symbol', None)
    cleaned['symbol'] = any_symbol(symbol)
    order_request = reduce_combine(request, cleaned)
    return func(order_request)



class _GlobalState:
    """
    Borg shared state class
    """
    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state
        if 'global_debugging' not in self.__shared_state:
            self.set_defaults()

    def set_defaults(self):
        self.raise_on_errors = False
        self.global_debugging = False
        self.max_bars = 100_000
        self.log = print

    @property
    def log(self):
        return self._log

    @log.setter
    def log(self, logger):
        self._log = logger or self._log


global_state: _GlobalState = _GlobalState()



import contextlib
import signal
import sys

from . import const
from . import helpers
from .core import account_info
from .core import initialize
from .core import last_error
from .core import MT5Error
from .core import shutdown
from .core import terminal_info
from .state import global_state as _state
from .types import *


@contextlib.contextmanager
def connected(*,
              path: str = None,
              portable: bool = None,
              server: str = None,
              login: int = None,
              password: str = None,
              timeout: int = None,
              ensure_trade_enabled: bool = False,
              enable_real_trading: bool = False,
              logger: Callable = None,
              raise_on_errors: bool = False,
              debug_logging: bool = False,
              # force_namedtuple: bool = False,
              **kwargs
              ) -> None:
    """Context manager for managing the connection with a MT5 terminal using the python ``with`` statement.

    >>> mt5_connected = mt5.connected(raise_on_errors=True, ensure_trade_enabled=True)
    >>> with mt5_connected:
    ...     print(mt5.symbols_get())
    >>>

    :param path:  Path to terminal.
    :param portable: Load terminal in portable mode.
    :param server:  Server name.
    :param login:  Account login number.
    :param password:  Account password.
    :param timeout: Connection init timeout.
    :param ensure_trade_enabled: Ensure that auto-trading is enabled. Will raise MT5Error when set to True and the terminal auto-trading is disabled.
    :param enable_real_trading:  Must be explicitly set to True to run on a live account.
    :param logger: Logging function. Will pass connection status messages to this function.
    :param raise_on_errors: Automatically checks last_error() after each function call and will raise a Mt5Error when the error-code is not RES_S_OK
    :param debug_logging: Logs each function call.

    :param kwargs:
    :return: None

    Note:
        The param ``enable_real_trading`` must be set to True to work on live accounts.
    """
    args = locals().copy()
    args = helpers.reduce_args(args)
    mt5_keys = ('path', 'portable', 'server', 'login', 'password', 'timeout')
    mt5_kwargs = {k: v for k, v in args.items() if k in mt5_keys}
    _state.global_debugging = bool(debug_logging)
    _state.raise_on_errors = bool(raise_on_errors)
    # _state.force_namedtuple = bool(force_namedtuple)
    _state.log = logger or print
    log = _state.log
    try:
        if not initialize(**mt5_kwargs):
            raise MT5Error(*last_error())
        elif debug_logging:
            log("MT5 connection has been initialized.")
        acc_info = account_info()
        if not enable_real_trading and acc_info.trade_mode == const.ACCOUNT_TRADE_MODE_REAL:
            raise MT5Error(
                const.RES_X_REAL_ACCOUNT_DISABLED,
                "REAL ACCOUNT TRADING HAS NOT BEEN ENABLED IN THE CONTEXT MANAGER")
        term_info = terminal_info()
        if ensure_trade_enabled and not term_info.trade_allowed:
            if debug_logging:
                log("Failed to initialize because auto-trade is disabled in terminal.")
            raise MT5Error(const.RES_X_AUTO_TRADE_DISABLED, "Terminal Auto-Trading is disabled.")
        _state.max_bars = term_info.maxbars
        yield
    finally:
        shutdown()
        if debug_logging:
            log("MT5 connection has been shutdown.")
        _state.set_defaults()


def _sigterm_handler(signum, frame):
    sys.exit(0)


_sigterm_handler.__enter_ctx__ = False


@contextlib.contextmanager
def handle_exit(callback=None, append=False):
    """A context manager which properly handles SIGTERM and SIGINT
    (KeyboardInterrupt) signals, registering a function which is
    guaranteed to be called after signals are received.
    Also, it makes sure to execute previously registered signal
    handlers as well (if any).

    >>> app = App()
    >>> with handle_exit(app.stop):
    ...     app.start()
    ...
    >>>

    If append == False raise RuntimeError if there's already a handler
    registered for SIGTERM, otherwise both new and old handlers are
    executed in this order.
    """
    old_handler = signal.signal(signal.SIGTERM, _sigterm_handler)
    if (old_handler != signal.SIG_DFL) and (old_handler != _sigterm_handler):
        if not append:
            raise RuntimeError("there is already a handler registered for "
                               "SIGTERM: %r" % old_handler)

        def handler(signum, frame):
            try:
                _sigterm_handler(signum, frame)
            finally:
                old_handler(signum, frame)

        signal.signal(signal.SIGTERM, handler)

    if _sigterm_handler.__enter_ctx__:
        raise RuntimeError("can't use nested contexts")
    _sigterm_handler.__enter_ctx__ = True

    try:
        yield
    except KeyboardInterrupt:
        pass
    except SystemExit as err:
        # code != 0 refers to an application error (e.g. explicit
        # sys.exit('some error') call).
        # We don't want that to pass silently.
        # Nevertheless, the 'finally' clause below will always
        # be executed.
        if err.code != 0:
            raise
    finally:
        _sigterm_handler.__enter_ctx__ = False
        if callback is not None:
            callback()
